// Copyright (C) MongoDB, Inc. 2017-present.
//
// Licensed under the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License. You may obtain
// a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

package benchmark

import "go.mongodb.org/mongo-driver/bson/primitive"

type flatBSONTags struct {
	ID primitive.ObjectID `bson:"_id"`

	AA  int64  `bson:"AAgSNVyBb"`
	AI  bool   `bson:"aicoMxZq"`
	AM  int64  `bson:"AMQrGQmu"`
	Ag  int    `bson:"AgYYbYPr"`
	Ah  int64  `bson:"ahFCBmqT"`
	At  int64  `bson:"AtWNZJXa"`
	BB  string `bson:"BBqZInWV"`
	BK  int64  `bson:"bkuaZWRT"`
	Bw  int    `bson:"BwTXiovJ"`
	CD  int    `bson:"CDIGOuIZ"`
	CEA string `bson:"CEtYKsdd"`
	CEB string `bson:"cepcgozk"`
	CF  int    `bson:"CFujXoob"`
	CV  int64  `bson:"cVjWCrlu"`
	CX  string `bson:"cxOHMeDJ"`
	CY  string `bson:"CYhSCkWB"`
	Cq  string `bson:"CqCssWxW"`
	DC  int    `bson:"dCLfYqqM"`
	DDA int    `bson:"ddPdLgGg"`
	DDB int    `bson:"ddVenEkK"`
	DH  string `bson:"dHsYhRbV"`
	DJ  int    `bson:"DJsnHZIC"`
	DN  string `bson:"dNSuxlSU"`
	DO  int64  `bson:"doshbrpF"`
	DP  string `bson:"dpbwfSRb"`
	DQ  int64  `bson:"DQBQcQFj"`
	DT  string `bson:"dtywOLeD"`
	DV  int    `bson:"dVkWIafN"`
	EG  bool   `bson:"egxZaSsw"`
	ER  string `bson:"eRTIdIJR"`
	FD  int64  `bson:"FDYGeSiR"`
	FE  string `bson:"fEheUtop"`
	Fp  bool   `bson:"FpduyhQP"`
	GE  string `bson:"gErhgZTh"`
	GY  int    `bson:"gySFZeAE"`
	Gi  uint   `bson:"GiAHzFII"`
	HN  string `bson:"hnVgYIQi"`
	HQA int    `bson:"HQeCoswW"`
	HQB int    `bson:"HQiykral"`
	HV  int64  `bson:"HVHyetUM"`
	HW  int    `bson:"hwHOTmmW"`
	Hi  bool   `bson:"HicJbMpj"`
	Hr  int    `bson:"HrUPbFHD"`
	IF  string `bson:"iFFGfTXc"`
	IJ  int    `bson:"ijwXMKqI"`
	IW  int    `bson:"iwfbMdcv"`
	Ib  string `bson:"Ibrdrtgg"`
	Is  bool   `bson:"IsorvnMR"`
	JB  string `bson:"jbUymqiB"`
	JM  string `bson:"jmglLvAS"`
	JW  int    `bson:"jWaFvVAz"`
	JX  int    `bson:"JXMyYkfb"`
	Jh  bool   `bson:"JhImQOkw"`
	Jr  string `bson:"JrJzKiIx"`
	Jz  int    `bson:"JzgaUWVG"`
	KF  bool   `bson:"kfvcFmKw"`
	KM  int64  `bson:"KMKBtlov"`
	Kn  string `bson:"KnhgtAOJ"`
	Ky  int    `bson:"KyxOoCqS"`
	LU  string `bson:"LUPqMOHS"`
	LV  bool   `bson:"LVNIFCYm"`
	Ln  int    `bson:"LngvlnTV"`
	ML  int    `bson:"mlfZVfVT"`
	MN  bool   `bson:"MNuWZMLP"`
	MX  int    `bson:"MXMxLVBk"`
	Mc  string `bson:"McpOBmaR"`
	Me  string `bson:"MeUYSkPS"`
	Mq  int    `bson:"MqfkBZJF"`
	NB  int    `bson:"nBKWWUWk"`
	NK  int    `bson:"nKhiSITP"`
	OB  int    `bson:"obcwwqWZ"`
	OC  string `bson:"OCsIhHxq"`
	OM  int    `bson:"omnwvBbA"`
	OR  string `bson:"oRWMNJTE"`
	Of  string `bson:"OfTmCvDx"`
	PA  int    `bson:"pacTBmxE"`
	PF  int    `bson:"PFZSRHNN"`
	PK  bool   `bson:"pKjOghFa"`
	PO  int    `bson:"pOMEwSod"`
	PP  string `bson:"pPtPsgRl"`
	PQ  int    `bson:"pQyCJaEd"`
	Pj  int    `bson:"PjKiuWnQ"`
	Pv  int    `bson:"PvfnpsMV"`
	QH  int    `bson:"qHzOMXeT"`
	QR  bool   `bson:"qrJASGzU"`
	Qo  string `bson:"QobifTeZ"`
	RE  int64  `bson:"reiKnuza"`
	RM  string `bson:"rmzUAgmk"`
	RP  string `bson:"RPsQhgRD"`
	Rb  uint   `bson:"Rbxpznea"`
	ReA bool   `bson:"RemSsnnR"`
	ReB int    `bson:"ReOZakjB"`
	Rw  string `bson:"RwAVVKHM"`
	SG  bool   `bson:"sGWJTAcT"`
	SU  uint8  `bson:"SUWXijHT"`
	SYA int64  `bson:"sYtnozSc"`
	SYB string `bson:"SYtZkQbC"`
	Sq  int64  `bson:"SqNvlUZF"`
	TA  int    `bson:"taoNnQYY"`
	TD  string `bson:"TDUzNJiH"`
	TI  string `bson:"tIJEYSYM"`
	TR  bool   `bson:"TRpgnInA"`
	Tg  int    `bson:"TgSwBbgp"`
	Tk  int64  `bson:"TkXMwZlU"`
	Tm  int64  `bson:"TmUnYUrv"`
	UK  int    `bson:"UKwbAKGw"`
	UM  string `bson:"uMDWqLMf"`
	Up  bool   `bson:"UpdMADoN"`
	Ut  int64  `bson:"UtbwOKLt"`
	VC  int64  `bson:"VCSKFCoE"`
	VK  string `bson:"vkEDWgmN"`
	VL  string `bson:"vlSZaxCV"`
	VS  string `bson:"vSLTtfDF"`
	VVA bool   `bson:"vvUeXASH"`
	VVB int    `bson:"VVvwKVRG"`
	Vc  bool   `bson:"VcCSqSmp"`
	Vp  int16  `bson:"VplFgewF"`
	Vt  string `bson:"VtzeOlCT"`
	WH  bool   `bson:"WHSQVLKG"`
	WJA bool   `bson:"wjfyueDC"`
	WJB string `bson:"wjAWaOog"`
	WM  int64  `bson:"wmDLUkXt"`
	WY  string `bson:"WYJdGJLu"`
	Wm  bool   `bson:"WmMOvgFc"`
	Wo  string `bson:"WoFGfdvb"`
	XE  int    `bson:"XEBqaXkB"`
	XG  bool   `bson:"XGxlHrXf"`
	XR  string `bson:"xrzGnsEK"`
	XWA int64  `bson:"xWpeGNjl"`
	XWB string `bson:"xWUlYggc"`
	XX  int64  `bson:"XXKbyIXG"`
	XZ  int64  `bson:"xZOksssj"`
	Xe  uint   `bson:"XeRkAyCq"`
	Xx  int    `bson:"XxvXmHiQ"`
	YD  string `bson:"YDHWnEXV"`
	YE  bool   `bson:"yeTUgNrU"`
	YK  int    `bson:"yKfZnGKG"`
	YX  string `bson:"yXSBbPeT"`
	ZD  bool   `bson:"zDzSGNnW"`
	ZE  bool   `bson:"zEgGhhZf"`
	ZM  string `bson:"zMCFzcWY"`
	ZSA int64  `bson:"zSYvADVf"`
	ZSB int64  `bson:"zswQbWEI"`
	Zm  string `bson:"ZmtEJFSO"`
}

type flatBSON struct {
	AMQrGQmu  int64
	AAgSNVyBb int64
	AgYYbYPr  int
	AtWNZJXa  int64
	BBqZInWV  string
	BwTXiovJ  int
	CDIGOuIZ  int
	CEtYKsdd  string
	CFujXoob  int
	CYhSCkWB  string
	CqCssWxW  string
	DJsnHZIC  int
	DQBQcQFj  int64
	FDYGeSiR  int64
	FpduyhQP  bool
	GiAHzFII  uint
	HQeCoswW  int
	HQiykral  int
	HVHyetUM  int64
	HicJbMpj  bool
	HrUPbFHD  int
	Ibrdrtgg  string
	IsorvnMR  bool
	JXMyYkfb  int
	JhImQOkw  bool
	JrJzKiIx  string
	JzgaUWVG  int
	KMKBtlov  int64
	KnhgtAOJ  string
	KyxOoCqS  int
	LUPqMOHS  string
	LVNIFCYm  bool
	LngvlnTV  int
	MNuWZMLP  bool
	MXMxLVBk  int
	McpOBmaR  string
	MeUYSkPS  string
	MqfkBZJF  int
	OCsIhHxq  string
	OfTmCvDx  string
	PjKiuWnQ  int
	PvfnpsMV  int
	QobifTeZ  string
	RPsQhgRD  string
	Rbxpznea  uint
	ReOZakjB  int
	RemSsnnR  bool
	RwAVVKHM  string
	SUWXijHT  uint8
	SYtZkQbC  string
	SqNvlUZF  int64
	TDUzNJiH  string
	TRpgnInA  bool
	TgSwBbgp  int
	TkXMwZlU  int64
	TmUnYUrv  int64
	UKwbAKGw  int
	UpdMADoN  bool
	UtbwOKLt  int64
	VCSKFCoE  int64
	VVvwKVRG  int
	VcCSqSmp  bool
	VplFgewF  int16
	VtzeOlCT  string
	WHSQVLKG  bool
	WYJdGJLu  string
	WmMOvgFc  bool
	WoFGfdvb  string
	XEBqaXkB  int
	XGxlHrXf  bool
	XXKbyIXG  int64
	XeRkAyCq  uint
	XxvXmHiQ  int
	YDHWnEXV  string
	ZmtEJFSO  string
	ID        primitive.ObjectID `bson:"_id"`
	AhFCBmqT  int64
	AicoMxZq  bool
	BkuaZWRT  int64
	CVjWCrlu  int64
	Cepcgozk  string
	CxOHMeDJ  string
	DCLfYqqM  int
	DHsYhRbV  string
	DNSuxlSU  string
	DVkWIafN  int
	DdPdLgGg  int
	DdVenEkK  int
	DoshbrpF  int64
	DpbwfSRb  string
	DtywOLeD  string
	ERTIdIJR  string
	EgxZaSsw  bool
	FEheUtop  string
	GErhgZTh  string
	GySFZeAE  int
	HnVgYIQi  string
	HwHOTmmW  int
	IFFGfTXc  string
	IjwXMKqI  int
	IwfbMdcv  int
	JWaFvVAz  int
	JbUymqiB  string
	JmglLvAS  string
	KfvcFmKw  bool
	MlfZVfVT  int
	NBKWWUWk  int
	NKhiSITP  int
	ORWMNJTE  string
	ObcwwqWZ  int
	OmnwvBbA  int
	PKjOghFa  bool
	POMEwSod  int
	PPtPsgRl  string
	PQyCJaEd  int
	PacTBmxE  int
	QHzOMXeT  int
	QrJASGzU  bool
	ReiKnuza  int64
	RmzUAgmk  string
	SGWJTAcT  bool
	SYtnozSc  int64
	TIJEYSYM  string
	TaoNnQYY  int
	UMDWqLMf  string
	VSLTtfDF  string
	VkEDWgmN  string
	VlSZaxCV  string
	VvUeXASH  bool
	WjAWaOog  string
	WjfyueDC  bool
	WmDLUkXt  int64
	XWUlYggc  string
	XWpeGNjl  int64
	XZOksssj  int64
	XrzGnsEK  string
	YKfZnGKG  int
	YXSBbPeT  string
	YeTUgNrU  bool
	ZDzSGNnW  bool
	ZEgGhhZf  bool
	ZMCFzcWY  string
	ZSYvADVf  int64
	ZswQbWEI  int64
	PfZSRHnn  int
}
